------------------------------------------------------------------------------
--  Ada95 Interface to Oracle RDBMS                                         --
--  Copyright (C) 2000-2001 Dmitriy Anisimkov.                              --
--  License text and authors contact information could be found in oci.ads  --
------------------------------------------------------------------------------
with
   System,
   Interfaces.C.Strings,
   Ada.Unchecked_Conversion,
--   Ada.Exceptions,
   Ada.Strings.Maps,
   Ada.Strings.Fixed,
   OCI.Thick.Strings,
   OCI.Thread;
--with Text_IO,System.Address_Image; use Text_IO,System;
package body OCI.Thick.Statements is
   use Lib;
   use type Sword;
   use type Ub4;
   use type OCIHandle;
   
   Dummy_Connect : Connection;

   use Ada.Strings.Maps;

   Dos_Fix_CR : constant Character_Mapping := To_Mapping (""&ASCII.Cr, " ");

   function to_Statement_Type is new Ada.Unchecked_Conversion (Integer, Statement_Type);

   -------------
   -- Destroy --
   -------------

   procedure Destroy (Object : in out Statement) is
   begin
      if Object.Handle /= Empty_Handle then
--         Put("Statement ");   
         OCI.Thick.Free(Object.Handle, OCI_HTYPE_STMT);
--         Put_Line("freed");   
      end if;
   end;

   function Get_Stmt_Attr (Stmt : Statement'Class; Attr : Ub4)
         return Integer is
      Result : aliased Integer := 0;
      Rsize : aliased Ub4 := Integer'Size/System.Storage_Unit;
      Rc : Sword := OCIAttrGet (Trgthndlp => Stmt.Handle,
         Trghndltyp => OCI_HTYPE_STMT,
         Attributep => Result'access,
         Sizep => Rsize'access,
         Attrtype => Attr,
         Errhp => Thread.Error);

   begin
      Check_Error(Rc);
      return Result;
   end Get_Stmt_Attr;

   function Number_Of_Columns(Stmt : Statement) return Natural
         is
   begin
      return Get_Stmt_Attr (Stmt, OCI_ATTR_Param_Count);
   end Number_Of_Columns;
   
   function Parse_Error_Offset (Stmt : in Statement) return Natural
   is   
   begin
      return Get_Stmt_Attr (Stmt, OCI_ATTR_PARSE_ERROR_OFFSET);
   end Parse_Error_Offset;


   function Type_Of_Statement(Stmt : Statement) return Statement_Type
         is
   begin
      return To_Statement_Type(Get_Stmt_Attr(Stmt, OCI_ATTR_STMT_TYPE));
   end Type_Of_Statement;

   -------------
   -- Prepare --
   -------------

   function Prepare (Connect : Connection; Code : String) return Statement
   is
      H : OCIHandle := Alloc_Handle (Thread.Environment, OCI_HTYPE_STMT);
      Rc : Sword;
   begin
      Rc := OCIStmtPrepare (OCIStmt(H),
         Thread.Error,
         C.To_C (Ada.Strings.Fixed.Translate (Code, Dos_Fix_CR)),
         Code'Length);
      Check_Error(Rc);
      return (RF.Controlled_Reference with Handle => H, 
               Connect=>Connect, 
               Executed => False,
               Described => False);
   end Prepare;

   function Prepare (Code : String) return Statement
   is
   begin
      return Prepare (Dummy_Connect, Code);
   end;

   --------------
   -- Describe --
   --------------

   procedure Execute (Connect : Connection; Stmt : in out Statement)
   is
   begin
      Stmt.Connect := Connect;
      Execute (Stmt);
   end;
   
   procedure Execute_Internal (Stmt : in out Statement;
      Raise_Exception : in out Boolean;
      Commit_On_Success : Boolean := False)
   is
      SType : Statement_Type := Type_Of_Statement (Stmt);
      Not_Select : boolean := Stype /= STMT_SELECT;
      Commit_Mode : constant array (boolean) of Ub4 := (False => OCI_DEFAULT, True => OCI_COMMIT_ON_SUCCESS);
      Rc : Sword := OCIStmtExecute (
         OCISvcCtx(Handle(Stmt.Connect)),
         OCIStmt(Stmt.Handle),
         Thread.Error,
         Boolean'Pos (Not_Select),
         Mode => Commit_Mode (Commit_On_Success));
   begin
      OCI.Thick.Strings.Complete_Last_Value;
      if Raise_Exception then
         Check_Error (Rc);
         Raise_Exception := False;
      else 
         if Rc = OCI_SUCCESS then
            Stmt.Executed := True;
            Stmt.Described := True;
         else
            Raise_Exception := True;
         end if;
      end if;
   end Execute_Internal;

   procedure Execute (Stmt : in out Statement) is
      Raise_Exception : Boolean := True;
   begin
      Execute_Internal (Stmt, Raise_Exception);
   end Execute;

   procedure Execute (Stmt : in out Statement; Success : out Boolean) is
      Raise_Exception : Boolean := False;
   begin
      Execute_Internal (Stmt, Raise_Exception => Raise_Exception);
      Success := not Raise_Exception;
   end Execute;

   procedure Execute_And_Commit (Stmt : in out Statement) is
      Raise_Exception : Boolean := True;
   begin
      Execute_Internal (Stmt,
         Raise_Exception => Raise_Exception,
         Commit_On_Success => True);
   end Execute_And_Commit;

   procedure Execute_And_Commit (Stmt : in out Statement; Success : out Boolean) is
      Raise_Exception : Boolean := False;
   begin
      Execute_Internal (Stmt,
         Raise_Exception => Raise_Exception,
         Commit_On_Success => True);
      Success := not Raise_Exception;
   end Execute_And_Commit;
   
   --------------
   -- Describe --
   --------------

   procedure Describe (Connect : Connection; Stmt : in out Statement)
   is
   begin
      Stmt.Connect := Connect;
      Describe (Stmt);
   end;

   procedure Describe (Stmt : in out Statement) is
      Rc : Sword := OCIStmtExecute (
         OCISvcCtx(Handle(Stmt.Connect)),
         OCIStmt(Stmt.Handle),
         Thread.Error,
         0,
         mode => OCI_DESCRIBE_ONLY);
   begin
      Check_Error (Rc);
      Stmt.Described := True;
   end Describe;
   
   procedure Describe (Stmt : in out Statement; Success : out Boolean)
   is
      Rc : Sword := OCIStmtExecute (
         OCISvcCtx(Handle(Stmt.Connect)),
         OCIStmt(Stmt.Handle),
         Thread.Error,
         0,
         mode => OCI_DESCRIBE_ONLY);
   begin
      Success := Rc = OCI_SUCCESS;
      if Success then
        Stmt.Described := True;
      end if;
   end Describe;
   
   procedure Set_Connection (Stmt : in out Statement; Connect : Connection) is
   begin
      Stmt.Connect := Connect;
   end;

   function Execute (Connect : Connection; Code : String) return Statement is
      Result : Statement := Prepare(Connect,Code);
   begin
      Execute (Result);
      return Result;
   end;
   
   procedure Execute (Connect : Connection; Code : String) is
     Dummy : Statement := Prepare(Connect,Code);
   begin
      Execute (Dummy);
   end;

   function Executed (Stmt : Statement) return boolean is
   begin
     return Stmt.Executed;
   end;

   function Described (Stmt : Statement) return boolean is
   begin
     return Stmt.Described;
   end;

   procedure Cancel (Stmt : Statement) is
      Rc : Sword := OCIStmtFetch (OCIStmt (Stmt.Handle),
         Thread.Error, Nrows => 0);
   begin
      Check_Error (Rc);
   end Cancel;

   function Fetch(Stmt : Statement) return Boolean is
      Rc : Sword := OCIStmtFetch (OCIStmt (Stmt.Handle),
         Thread.Error);
   begin
      OCI.Thick.Strings.Complete_Last_Value;
      case RC is
         when OCI_NO_DATA => return False;
         when OCI_SUCCESS => return True;
         when others => Check_Error (Rc);
      end case;
      --  We should not be there.
      raise PROGRAM_ERROR;
   end Fetch;

   function Get_Connection(Stmt : Statement) return Connection is
   begin
     return Stmt.Connect;
   end;
   
   function Bind_Names(Stmt : Statement) return VStrings is
      Names : aliased Vector_Of_OraText (1 .. 256);
      Lens : aliased vector_of_ub1 (Names'Range);
      INames : aliased Vector_Of_OraText (Names'Range);
      ILens : aliased vector_of_ub1 (Names'Range);
      Dups : aliased vector_of_ub1 (Names'Range);
      Binds : aliased vector_of_OCIBIND (Names'Range);
      Found : aliased sb4;
      rc : sword := OCISTMTGETBINDINFO (stmtp  => OCISTMT(Stmt.Handle),
                                errhp   => Thread.Error,
                                size    => Names'Length,
                                startloc=> 1,
                                found   => Found'Unchecked_Access,
                                bvnp    => Names (Names'First)'Unchecked_Access,
                                bvnl    => Lens (Lens'First)'Unchecked_Access,
                                invp    => INames (INames'First)'Unchecked_Access,
                                inpl    => ILens (ILens'First)'Unchecked_Access,
                                dupl    => Dups (Dups'First)'Unchecked_Access,
                                hndl    => Binds (Binds'First)'Unchecked_Access);
     use Interfaces.C.Strings;
     use type c.size_t;
     use type ub1;
   begin
     if RC = OCI_NO_DATA then
        return (1 .. 0 => NULL_Unbounded_String);
     end if;
     Check_Error(rc);
     declare
       Result : VStrings (1 .. Natural (Found));
       count : Natural := 0;
     begin
       for i in Result'Range loop
          if Dups(i) = 0 then
            Count := Count+1;
            Result (Count) := To_Unbounded_String (
              C.To_Ada (Value (Names (i), c.size_t (Lens (i))), False));
          end if;  
--          Put_Line(C.To_Ada(Value(Names(i),c.size_t(Lens(i))),False)&
--             ' '&C.To_Ada(Value(INames(i),c.size_t(ILens(i))),False));
       end loop; 
       return Result (1 .. Count);
     end;  
   end;

end OCI.Thick.Statements;

