------------------------------------------------------------------------------
--  Thin Ada95 binding to OCI (Oracle Call Interface)                    --
--  Copyright (C) 2000-2002 Dmitriy Anisimkov.                              --
--  License agreement and authors contact information are in file oci.ads   --
------------------------------------------------------------------------------

--  $Id$

with
   Ada.Exceptions,
   OCI.Thread,
   OCI.Environments;
with Ada.Text_IO;
--with System.Address_Image; use Ada.Text_IO,System;
package body Oci.Thick is
   use Lib;
   use type Sword;
   use type Ub4;
   use type OCIHandle;
   
-- Local utilities

   procedure Task_Dedicated_Environment is
   begin
      if not Thread.Synch.Task_Dedicated_Environment then
         raise Too_Late_For_Dedicated_Environment;
      end if;
   end;
   
   procedure Task_Done (T : Task_ID := Current_Task)
   is
   begin
      Thread.Attributes.Reinitialize (T);
   end;  

   -- Should call before first OCI call.
   procedure Set_Events_Support
   is
   begin
      OCI.Environments.Set_Create_Mode_Flag (OCI_EVENTS);
   end;
   
   procedure Set_Objects_Support
   is
   begin
      OCI.Environments.Set_Create_Mode_Flag (OCI_OBJECT);
   end;
 
   function Last_Error_Code return Integer
   is
      Rc : Sword;
      Errcodep : aliased Sb4 := 0;
   begin
       Rc := OCIErrorGet(Hndlp => Thread.Error,
               Errcodep => Errcodep'access,
               Bufp => CStr.null_ptr,
               Bufsiz => 0,
               Htype => OCI_HTYPE_ERROR);
       return Integer(Errcodep);
   end;
   
   function Last_Error_Message (Record_No : Integer := 1) return String
   is
      Rc : Sword;
      Errcodep : aliased Sb4 := 0;
      Bufp : aliased C.Char_Array := (0 .. 4095 => C.nul);
   begin
      Rc := OCIErrorGet (Hndlp => Thread.Error,
               Errcodep => Errcodep'access,
               Recordno => Ub4 (Record_No),
               Bufp => CStr.To_Chars_Ptr(Bufp'Unchecked_Access),
               Bufsiz => Bufp'Length-1,
               Htype => OCI_HTYPE_ERROR);
      return C.To_Ada(Bufp);
   end;
 
   function Last_Error_Offset return Natural
   is
   begin
      return Get_Attr (OCIHandle (Thread.Error),
         OCI_HTYPE_ERROR,
         OCI_ATTR_DML_ROW_OFFSET);
   end Last_Error_Offset;

   procedure Print_Error (Msg : String)
   is
      use Ada.Text_IO;
   begin
      Put_Line (Current_Error, Msg);
   end;
      

   procedure Check_Error (
         Code : Sword;
         Handle : OCIError;
         Htype : Integer; Raise_Error : Boolean)
   is
      use Ada.Exceptions;
      Bufp : aliased C.Char_Array := (0 .. 2047 => C.nul);
      Errcodep : aliased Sb4 := 0;
      Rc : Sword;
      
      procedure Error (Except : Exception_Id; Msg : String)
      is
      begin
         if Raise_Error then
            Raise_Exception (Except, Msg);
         else
            Print_Error (Msg);
         end if;
      end;
      
   begin
      case Code is
         when OCI_ERROR | OCI_NO_DATA | OCI_SUCCESS_WITH_INFO =>
            Rc := OCIErrorGet
              (Hndlp => Handle,
               Errcodep => Errcodep'access,
               Bufp => CStr.To_Chars_Ptr (Bufp'Unchecked_Access),
               Bufsiz => Bufp'Length - 1,
               Htype => UB4 (Htype));
            if Rc = OCI_SUCCESS then
               Error (LIB_ERROR'Identity, C.To_Ada (Bufp));
            else
               Error
                 (LIB_ERROR'Identity,
                  "Error code:" & Sword'Image(Code) & Ascii.Lf &
                  "Return error code:" & Sword'Image (Rc) & Ascii.Lf &
                  "Output error code:" & Sb4'Image (Errcodep) & Ascii.Lf &
                  "Message:" & C.To_Ada (Bufp));
            end if;
         when OCI_INVALID_HANDLE =>
            Error (INVALID_HANDLE'Identity, "Invalid handle");
         when OCI_SUCCESS => null;
         when others =>
            Error (CONSTRAINT_ERROR'Identity, "Error code" & SWord'Image (Code));
      end case;
   end Check_Error;

   procedure Check_Error (
         Code : Sword; Raise_Exception : Boolean := True) is
   begin
      Check_Error(Code, Thread.Error,
         OCI_HTYPE_ERROR, Raise_Exception);
   end Check_Error;

   function Alloc_Handle (Parent : OCIEnv; Htype : Ub4) return OCIHandle
         is
      Result : aliased OCIHandle := Empty_Handle;
      Rc : Sword := OCIHandleAlloc(Parenth => OCIHandle(Parent),
         Hndlpp => Result'access,
         Htype => Htype);
   begin
      if Rc /= OCI_SUCCESS then
         raise INVALID_HANDLE;
      end if;
      return Result;
   end;

   function Alloc_Descriptor (Parent : OCIEnv; Htype : Ub4) return OCIHandle
   is
      Result : aliased OCIHandle := Empty_Handle;
      Rc : Sword := OCIDescriptorAlloc(Parenth => Parent,
         Descpp => Result'access,
         Htype => Htype);
   begin
      if Rc /= OCI_SUCCESS then
         raise INVALID_HANDLE;
      end if;
      return Result;
   end;

   procedure Free (H : in out OCIHandle; HType : Ub4) is
      Rc : Sword;
   begin
      if H = Empty_Handle then
         return;
      end if;
      RC := OCIHandleFree (H, HType);
      H := Empty_Handle;
      Check_Error (RC, Raise_Exception => False);
   end Free;
   
   function Get_Attr (Param : OCIHandle; HType : ub4; Attr : Ub4) return String
         is
      Buff : aliased Cstr.Chars_Ptr;
      Rsize : aliased Ub4;
      Rc : Sword := OCIAttrGet (Trgthndlp => Param,
         Trghndltyp => HType,
         Attributep => Buff'access,
         Sizep => Rsize'access,
         Attrtype => Attr,
         Errhp => Thread.Error);
       
   begin
      Check_Error(Rc);
      if cstr."=" (Buff, cstr.Null_Ptr) then
         return "";
      else
         return C.To_Ada (Cstr.Value (Buff, C.Size_T (Rsize)), False);
      end if;  
   end Get_Attr;

   function Get_Attr (Param : OCIHandle; HType : ub4; Attr : Ub4) return Integer
         is
      Result : aliased Integer := 0;
      Rsize : aliased Ub4;
      Rc : Sword := OCIAttrGet (Trgthndlp => Param,
         Trghndltyp => HType,
         Attributep => Result'access,
         Sizep => Rsize'access,
         Attrtype => Attr,
         Errhp => Thread.Error);
   begin
      Check_Error (Rc);
      return Result;
   end Get_Attr;

   procedure Set_Attr (H : OCIHandle; HType : ub4; Value : String; Attr : Ub4)
   is
      rc : SWord := OCIAttrSet (Trgthndlp => H,
         Trghndltyp => HType,
         Attributep => C.To_C (Value),
         Size => Value'Length,
         Attrtype => Attr,
         Errhp => Thread.Error);
   begin
      Check_Error (Rc);
   end Set_Attr;


   procedure Set_Attr (H : OCIHandle; HType : ub4; Value : OCIHandle; Attr : Ub4)
   is
      rc : SWord := OCIAttrSet (Trgthndlp => H,
         Trghndltyp => HType,
         Attributep => Value,
         Size => 0,
         Attrtype => Attr,
         Errhp => Thread.Error);
   begin
      Check_Error (Rc);
   end Set_Attr;

-- Base_Variable
   
   procedure Clear_Value (Var : in out Limited_Variable'Class) is
   begin
      Var.Indicator := Null_Indicator;
   end;

   function Is_Null (Var : Limited_Variable'Class) return Boolean is
      use type Sb2;
   begin
      if Var.Define=OCIDefine(Empty_Handle)
            and Var.Bind=OCIBind(Empty_Handle) then
         raise NOT_ATTACHED;
      end if;
      return Var.Indicator = Null_Indicator;
   end;

   function Handle(ref : Handle_REference'Class) return OCIHandle is
   begin
     return ref.Handle;
   end;
 
end Oci.Thick;